@@ -6,20 +6,23 @@ require "zlib"
 require "set"
 require "active_support/dependencies"
 require "active_support/core_ext/digest/uuid"
-require "active_record/fixture_set/file"
-require "active_record/fixture_set/render_context"
-require "active_record/fixture_set/table_rows"
 require "active_record/test_fixtures"
 
 module ActiveRecord
   class FixtureClassNotFound < ActiveRecord::ActiveRecordError # :nodoc:
   end
 
+  # = Active Record \Fixtures
+  #
   # \Fixtures are a way of organizing data that you want to test against; in short, sample data.
   #
-  # They are stored in YAML files, one file per model, which are placed in the directory
-  # appointed by <tt>ActiveSupport::TestCase.fixture_path=(path)</tt> (this is automatically
-  # configured for Rails, so you can just put your files in <tt><your-rails-app>/test/fixtures/</tt>).
+  # They are stored in YAML files, one file per model, which are by default placed in either
+  # <tt><your-rails-app>/test/fixtures/</tt> or in the <tt>test/fixtures</tt>
+  # folder under any of your application's engines.
+  #
+  # The location can also be changed with ActiveSupport::TestCase.fixture_paths=,
+  # once you have <tt>require "rails/test_help"</tt> in your +test_helper.rb+.
+  #
   # The fixture file ends with the +.yml+ file extension, for example:
   # <tt><your-rails-app>/test/fixtures/web_sites.yml</tt>).
   #
@@ -39,10 +42,17 @@ module ActiveRecord
   # is followed by an indented list of key/value pairs in the "key: value" format. Records are
   # separated by a blank line for your viewing pleasure.
   #
-  # Note: Fixtures are unordered. If you want ordered fixtures, use the omap YAML type.
-  # See https://yaml.org/type/omap.html
-  # for the specification. You will need ordered fixtures when you have foreign key constraints
-  # on keys in the same table. This is commonly needed for tree structures. Example:
+  # == Ordering
+  #
+  # Fixtures by default are unordered. This is because the maps in YAML are unordered.
+  #
+  # If you want ordered fixtures, use the omap YAML type.
+  # See https://yaml.org/type/omap.html for the specification.
+  #
+  # You will need ordered fixtures when you have foreign key constraints
+  # on keys in the same table. This is commonly needed for tree structures.
+  #
+  # For example:
   #
   #    --- !omap
   #    - parent:
@@ -54,7 +64,7 @@ module ActiveRecord
   #        parent_id:  1
   #        title:      Child
   #
-  # = Using Fixtures in Test Cases
+  # == Using Fixtures in Test Cases
   #
   # Since fixtures are a testing construct, we use them in our unit and functional tests. There
   # are two ways to use the fixtures, but first let's take a look at a sample unit test:
@@ -100,6 +110,12 @@ module ActiveRecord
   #     assert_raise(StandardError) { web_sites(:reddit) }
   #   end
   #
+  # If the model names conflicts with a +TestCase+ methods, you can use the generic +fixture+ accessor
+  #
+  #   test "generic find" do
+  #     assert_equal "Ruby on Rails", fixture(:web_sites, :rubyonrails).name
+  #   end
+  #
   # Alternatively, you may enable auto-instantiation of the fixture data. For instance, take the
   # following tests:
   #
@@ -124,7 +140,7 @@ module ActiveRecord
   # traversed in the database to create the fixture hash and/or instance variables. This is expensive for
   # large sets of fixtured data.
   #
-  # = Dynamic fixtures with ERB
+  # == Dynamic fixtures with \ERB
   #
   # Sometimes you don't care about the content of the fixtures as much as you care about the volume.
   # In these cases, you can mix ERB in with your YAML fixtures to create a bunch of fixtures for load
@@ -162,7 +178,7 @@ module ActiveRecord
   #       name: kitten.png
   #       sha: <%= file_sha 'files/kitten.png' %>
   #
-  # = Transactional Tests
+  # == Transactional Tests
   #
   # Test cases can use begin+rollback to isolate their changes to the database instead of having to
   # delete+insert for every test case.
@@ -198,7 +214,7 @@ module ActiveRecord
   # 2. Your database does not support transactions. Every Active Record database supports transactions except MySQL MyISAM.
   #    Use InnoDB, MaxDB, or NDB instead.
   #
-  # = Advanced Fixtures
+  # == Advanced Fixtures
   #
   # Fixtures that don't specify an ID get some extra features:
   #
@@ -212,7 +228,7 @@ module ActiveRecord
   # * Fixture label interpolation
   # * Support for YAML defaults
   #
-  # == Stable, Autogenerated IDs
+  # === Stable, Autogenerated IDs
   #
   # Here, have a monkey fixture:
   #
@@ -241,13 +257,13 @@ module ActiveRecord
   # The generated ID for a given label is constant, so we can discover
   # any fixture's ID without loading anything, as long as we know the label.
   #
-  # == Label references for associations (+belongs_to+, +has_one+, +has_many+)
+  # === Label references for associations (+belongs_to+, +has_one+, +has_many+)
   #
   # Specifying foreign keys in fixtures can be very fragile, not to
   # mention difficult to read. Since Active Record can figure out the ID of
   # any fixture from its label, you can specify FK's by label instead of ID.
   #
-  # === +belongs_to+
+  # ==== +belongs_to+
   #
   # Let's break out some more monkeys and pirates.
   #
@@ -258,6 +274,8 @@ module ActiveRecord
   #     name: Reginald the Pirate
   #     monkey_id: 1
   #
+  # <code></code>
+  #
   #   ### in monkeys.yml
   #
   #   george:
@@ -275,6 +293,8 @@ module ActiveRecord
   #     name: Reginald the Pirate
   #     monkey: george
   #
+  # <code></code>
+  #
   #   ### in monkeys.yml
   #
   #   george:
@@ -296,6 +316,8 @@ module ActiveRecord
   #
   #   belongs_to :eater, polymorphic: true
   #
+  # <code></code>
+  #
   #   ### in fruits.yml
   #
   #   apple:
@@ -311,9 +333,9 @@ module ActiveRecord
   #
   # Just provide the polymorphic target type and Active Record will take care of the rest.
   #
-  # === +has_and_belongs_to_many+ or <tt>has_many :through</tt>
+  # ==== +has_and_belongs_to_many+ or <tt>has_many :through</tt>
   #
-  # Time to give our monkey some fruit.
+  # \Time to give our monkey some fruit.
   #
   #   ### in monkeys.yml
   #
@@ -321,6 +343,8 @@ module ActiveRecord
   #     id: 1
   #     name: George the Monkey
   #
+  # <code></code>
+  #
   #   ### in fruits.yml
   #
   #   apple:
@@ -335,6 +359,8 @@ module ActiveRecord
   #     id: 3
   #     name: grape
   #
+  # <code></code>
+  #
   #   ### in fruits_monkeys.yml
   #
   #   apple_george:
@@ -358,6 +384,8 @@ module ActiveRecord
   #     name: George the Monkey
   #     fruits: apple, orange, grape
   #
+  # <code></code>
+  #
   #   ### in fruits.yml
   #
   #   apple:
@@ -375,7 +403,7 @@ module ActiveRecord
   # the fixture's model class and discovers the +has_and_belongs_to_many+
   # associations.
   #
-  # == Autofilled Timestamp Columns
+  # === Autofilled \Timestamp Columns
   #
   # If your table/model specifies any of Active Record's
   # standard timestamp columns (+created_at+, +created_on+, +updated_at+, +updated_on+),
@@ -383,7 +411,7 @@ module ActiveRecord
   #
   # If you've set specific values, they'll be left alone.
   #
-  # == Fixture label interpolation
+  # === Fixture label interpolation
   #
   # The label of the current fixture is always available as a column value:
   #
@@ -400,7 +428,11 @@ module ActiveRecord
   #     monkey_id: <%= ActiveRecord::FixtureSet.identify(:reginald) %>
   #     pirate_id: <%= ActiveRecord::FixtureSet.identify(:george) %>
   #
-  # == Support for YAML defaults
+  # If the model uses UUID values for identifiers, add the +:uuid+ argument:
+  #
+  #   ActiveRecord::FixtureSet.identify(:boaty_mcboatface, :uuid)
+  #
+  # === Support for YAML defaults
   #
   # You can set and reuse defaults in your fixtures YAML file.
   # This is the same technique used in the +database.yml+ file to specify
@@ -442,6 +474,45 @@ module ActiveRecord
   # In the above example, 'base' will be ignored when creating fixtures.
   # This can be used for common attributes inheriting.
   #
+  # == Composite Primary Key Fixtures
+  #
+  # Fixtures for composite primary key tables are fairly similar to normal tables.
+  # When using an id column, the column may be omitted as usual:
+  #
+  #   # app/models/book.rb
+  #   class Book < ApplicationRecord
+  #     self.primary_key = [:author_id, :id]
+  #     belongs_to :author
+  #   end
+  #
+  # <code></code>
+  #
+  #   # books.yml
+  #   alices_adventure_in_wonderland:
+  #     author_id: <%= ActiveRecord::FixtureSet.identify(:lewis_carroll) %>
+  #     title: "Alice's Adventures in Wonderland"
+  #
+  # However, in order to support composite primary key relationships,
+  # you must use the `composite_identify` method:
+  #
+  #   # app/models/book_orders.rb
+  #   class BookOrder < ApplicationRecord
+  #     self.primary_key = [:shop_id, :id]
+  #     belongs_to :order, foreign_key: [:shop_id, :order_id]
+  #     belongs_to :book, foreign_key: [:author_id, :book_id]
+  #   end
+  #
+  # <code></code>
+  #
+  #   # book_orders.yml
+  #   alices_adventure_in_wonderland_in_books:
+  #     author: lewis_carroll
+  #     book_id: <%= ActiveRecord::FixtureSet.composite_identify(
+  #                  :alices_adventure_in_wonderland, Book.primary_key)[:id] %>
+  #     shop: book_store
+  #     order_id: <%= ActiveRecord::FixtureSet.composite_identify(
+  #                  :books, Order.primary_key)[:id] %>
+  #
   # == Configure the fixture model class
   #
   # It's possible to set the fixture's model class directly in the YAML file.
@@ -456,6 +527,10 @@ module ActiveRecord
   #
   # Any fixtures labeled "_fixture" are safely ignored.
   class FixtureSet
+    require "active_record/fixture_set/file"
+    require "active_record/fixture_set/render_context"
+    require "active_record/fixture_set/table_rows"
+
     #--
     # An instance of FixtureSet is normally stored in a single YAML file and
     # possibly in a folder with the same name.
@@ -467,39 +542,6 @@ module ActiveRecord
 
     cattr_accessor :all_loaded_fixtures, default: {}
 
-    class ClassCache
-      def initialize(class_names, config)
-        @class_names = class_names.stringify_keys
-        @config      = config
-
-        # Remove string values that aren't constants or subclasses of AR
-        @class_names.delete_if do |klass_name, klass|
-          !insert_class(@class_names, klass_name, klass)
-        end
-      end
-
-      def [](fs_name)
-        @class_names.fetch(fs_name) do
-          klass = default_fixture_model(fs_name, @config).safe_constantize
-          insert_class(@class_names, fs_name, klass)
-        end
-      end
-
-      private
-        def insert_class(class_names, name, klass)
-          # We only want to deal with AR objects.
-          if klass && klass < ActiveRecord::Base
-            class_names[name] = klass
-          else
-            class_names[name] = nil
-          end
-        end
-
-        def default_fixture_model(fs_name, config)
-          ActiveRecord::FixtureSet.default_fixture_model_name(fs_name, config)
-        end
-    end
-
     class << self
       def default_fixture_model_name(fixture_set_name, config = ActiveRecord::Base) # :nodoc:
         config.pluralize_table_names ?
@@ -517,24 +559,24 @@ module ActiveRecord
         @@all_cached_fixtures.clear
       end
 
-      def cache_for_connection(connection)
-        @@all_cached_fixtures[connection]
+      def cache_for_connection_pool(connection_pool)
+        @@all_cached_fixtures[connection_pool]
       end
 
-      def fixture_is_cached?(connection, table_name)
-        cache_for_connection(connection)[table_name]
+      def fixture_is_cached?(connection_pool, table_name)
+        cache_for_connection_pool(connection_pool)[table_name]
       end
 
-      def cached_fixtures(connection, keys_to_fetch = nil)
+      def cached_fixtures(connection_pool, keys_to_fetch = nil)
         if keys_to_fetch
-          cache_for_connection(connection).values_at(*keys_to_fetch)
+          cache_for_connection_pool(connection_pool).values_at(*keys_to_fetch)
         else
-          cache_for_connection(connection).values
+          cache_for_connection_pool(connection_pool).values
         end
       end
 
-      def cache_fixtures(connection, fixtures_map)
-        cache_for_connection(connection).update(fixtures_map)
+      def cache_fixtures(connection_pool, fixtures_map)
+        cache_for_connection_pool(connection_pool).update(fixtures_map)
       end
 
       def instantiate_fixtures(object, fixture_set, load_instances = true)
@@ -552,31 +594,30 @@ module ActiveRecord
         end
       end
 
-      def create_fixtures(fixtures_directory, fixture_set_names, class_names = {}, config = ActiveRecord::Base, &block)
+      def create_fixtures(fixtures_directories, fixture_set_names, class_names = {}, config = ActiveRecord::Base)
         fixture_set_names = Array(fixture_set_names).map(&:to_s)
-        class_names = ClassCache.new class_names, config
-
-        # FIXME: Apparently JK uses this.
-        connection = block_given? ? block : lambda { ActiveRecord::Base.connection }
+        class_names.stringify_keys!
 
+        connection_pool = config.connection_pool
         fixture_files_to_read = fixture_set_names.reject do |fs_name|
-          fixture_is_cached?(connection.call, fs_name)
+          fixture_is_cached?(connection_pool, fs_name)
         end
 
         if fixture_files_to_read.any?
           fixtures_map = read_and_insert(
-            fixtures_directory,
+            Array(fixtures_directories),
             fixture_files_to_read,
             class_names,
-            connection,
+            connection_pool,
           )
-          cache_fixtures(connection.call, fixtures_map)
+          cache_fixtures(connection_pool, fixtures_map)
         end
-        cached_fixtures(connection.call, fixture_set_names)
+        cached_fixtures(connection_pool, fixture_set_names)
       end
 
       # Returns a consistent, platform-independent identifier for +label+.
-      # Integer identifiers are values less than 2^30. UUIDs are RFC 4122 version 5 SHA-1 hashes.
+      #
+      # \Integer identifiers are values less than 2^30. UUIDs are RFC 4122 version 5 SHA-1 hashes.
       def identify(label, column_type = :integer)
         if column_type == :uuid
           Digest::UUID.uuid_v5(Digest::UUID::OID_NAMESPACE, label.to_s)
@@ -585,40 +626,54 @@ module ActiveRecord
         end
       end
 
-      # Superclass for the evaluation contexts used by ERB fixtures.
+      # Returns a consistent, platform-independent hash representing a mapping
+      # between the label and the subcomponents of the provided composite key.
+      #
+      # Example:
+      #
+      #   composite_identify("label", [:a, :b, :c]) # => { a: hash_1, b: hash_2, c: hash_3 }
+      def composite_identify(label, key)
+        key
+          .index_with
+          .with_index { |sub_key, index| (identify(label) << index) % MAX_ID }
+          .with_indifferent_access
+      end
+
+      # Superclass for the evaluation contexts used by \ERB fixtures.
       def context_class
         @context_class ||= Class.new
       end
 
       private
-        def read_and_insert(fixtures_directory, fixture_files, class_names, connection) # :nodoc:
+        def read_and_insert(fixtures_directories, fixture_files, class_names, connection_pool) # :nodoc:
           fixtures_map = {}
+          directory_glob = "{#{fixtures_directories.join(",")}}"
           fixture_sets = fixture_files.map do |fixture_set_name|
             klass = class_names[fixture_set_name]
             fixtures_map[fixture_set_name] = new( # ActiveRecord::FixtureSet.new
               nil,
               fixture_set_name,
               klass,
-              ::File.join(fixtures_directory, fixture_set_name)
+              ::File.join(directory_glob, fixture_set_name)
             )
           end
           update_all_loaded_fixtures(fixtures_map)
 
-          insert(fixture_sets, connection)
+          insert(fixture_sets, connection_pool)
 
           fixtures_map
         end
 
-        def insert(fixture_sets, connection) # :nodoc:
-          fixture_sets_by_connection = fixture_sets.group_by do |fixture_set|
+        def insert(fixture_sets, connection_pool) # :nodoc:
+          fixture_sets_by_pool = fixture_sets.group_by do |fixture_set|
             if fixture_set.model_class
-              fixture_set.model_class.connection
+              fixture_set.model_class.connection_pool
             else
-              connection.call
+              connection_pool
             end
           end
 
-          fixture_sets_by_connection.each do |conn, set|
+          fixture_sets_by_pool.each do |pool, set|
             table_rows_for_connection = Hash.new { |h, k| h[k] = [] }
 
             set.each do |fixture_set|
@@ -627,19 +682,29 @@ module ActiveRecord
               end
             end
 
-            conn.insert_fixtures_set(table_rows_for_connection, table_rows_for_connection.keys)
+            pool.with_connection do |conn|
+              conn.insert_fixtures_set(table_rows_for_connection, table_rows_for_connection.keys)
 
-            if ActiveRecord.verify_foreign_keys_for_fixtures && !conn.all_foreign_keys_valid?
-              raise "Foreign key violations found in your fixture data. Ensure you aren't referring to labels that don't exist on associations."
-            end
+              check_all_foreign_keys_valid!(conn)
 
-            # Cap primary key sequences to max(pk).
-            if conn.respond_to?(:reset_pk_sequence!)
-              set.each { |fs| conn.reset_pk_sequence!(fs.table_name) }
+              # Cap primary key sequences to max(pk).
+              if conn.respond_to?(:reset_pk_sequence!)
+                set.each { |fs| conn.reset_pk_sequence!(fs.table_name) }
+              end
             end
           end
         end
 
+        def check_all_foreign_keys_valid!(conn)
+          return unless ActiveRecord.verify_foreign_keys_for_fixtures
+
+          begin
+            conn.check_all_foreign_keys_valid!
+          rescue ActiveRecord::StatementInvalid => e
+            raise "Foreign key violations found in your fixture data. Ensure you aren't referring to labels that don't exist on associations. Error from database:\n\n#{e.message}"
+          end
+        end
+
         def update_all_loaded_fixtures(fixtures_map) # :nodoc:
           all_loaded_fixtures.update(fixtures_map)
         end
@@ -653,7 +718,6 @@ module ActiveRecord
       @config   = config
 
       self.model_class = class_name
-
       @fixtures = read_fixture_files(path)
 
       @table_name = model_class&.table_name || self.class.default_fixture_table_name(name, config)
@@ -715,14 +779,18 @@ module ActiveRecord
       # Loads the fixtures from the YAML file at +path+.
       # If the file sets the +model_class+ and current instance value is not set,
       # it uses the file value.
+
       def read_fixture_files(path)
-        yaml_files = Dir["#{path}/{**,*}/*.yml"].select { |f|
+        yaml_files = Dir["#{path}{.yml,/{**,*}/*.yml}"].select { |f|
           ::File.file?(f)
-        } + [yaml_file_path(path)]
+        }
+
+        raise ArgumentError, "No fixture files found for #{@name}" if yaml_files.empty?
 
         yaml_files.each_with_object({}) do |file, fixtures|
           FixtureSet::File.open(file) do |fh|
             self.model_class ||= fh.model_class if fh.model_class
+            self.model_class ||= default_fixture_model_class
             self.ignored_fixtures ||= fh.ignored_fixtures
             fh.each do |fixture_name, row|
               fixtures[fixture_name] = ActiveRecord::Fixture.new(row, model_class)
@@ -731,8 +799,9 @@ module ActiveRecord
         end
       end
 
-      def yaml_file_path(path)
-        "#{path}.yml"
+      def default_fixture_model_class
+        klass = ActiveRecord::FixtureSet.default_fixture_model_name(@name, @config).safe_constantize
+        klass if klass && klass < ActiveRecord::Base
       end
   end
 
@@ -769,7 +838,8 @@ module ActiveRecord
     def find
       raise FixtureClassNotFound, "No class attached to find." unless model_class
       object = model_class.unscoped do
-        model_class.find(fixture[model_class.primary_key])
+        pk_clauses = fixture.slice(*Array(model_class.primary_key))
+        model_class.find_by!(pk_clauses)
       end
       # Fixtures can't be eagerly loaded
       object.instance_variable_set(:@strict_loading, false)
